@@ -1,13 +1,3 @@
-#include <sys/types.h>
-
-#include <unistd.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <string.h>
-#include <stdlib.h>
-#include <ctype.h>
-#include <assert.h>
-
 #include "buffer.h"
 #include "server.h"
 #include "keyvalue.h"
@@ -24,6 +14,16 @@
 #include "inet_ntop_cache.h"
 #include "crc32.h"
 
+#include <sys/types.h>
+
+#include <unistd.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <string.h>
+#include <stdlib.h>
+#include <ctype.h>
+#include <assert.h>
+
 #include <stdio.h>
 
 #ifdef HAVE_SYS_FILIO_H
@@ -332,7 +332,7 @@ SETDEFAULTS_FUNC(mod_proxy_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-void proxy_connection_close(server *srv, handler_ctx *hctx) {
+static void proxy_connection_close(server *srv, handler_ctx *hctx) {
 	plugin_data *p;
 	connection *con;
 
@@ -356,20 +356,35 @@ void proxy_connection_close(server *srv, handler_ctx *hctx) {
 static int proxy_establish_connection(server *srv, handler_ctx *hctx) {
 	struct sockaddr *proxy_addr;
 	struct sockaddr_in proxy_addr_in;
+#if defined(HAVE_IPV6) && defined(HAVE_INET_PTON)
+	struct sockaddr_in6 proxy_addr_in6;
+#endif
 	socklen_t servlen;
 
 	plugin_data *p    = hctx->plugin_data;
 	data_proxy *host= hctx->host;
 	int proxy_fd       = hctx->fd;
 
-	memset(&proxy_addr, 0, sizeof(proxy_addr));
 
-	proxy_addr_in.sin_family = AF_INET;
-	proxy_addr_in.sin_addr.s_addr = inet_addr(host->host->ptr);
-	proxy_addr_in.sin_port = htons(host->port);
-	servlen = sizeof(proxy_addr_in);
+#if defined(HAVE_IPV6) && defined(HAVE_INET_PTON)
+	if (strstr(host->host->ptr, ":")) {
+		memset(&proxy_addr_in6, 0, sizeof(proxy_addr_in6));
+		proxy_addr_in6.sin6_family = AF_INET6;
+		inet_pton(AF_INET6, host->host->ptr, (char *) &proxy_addr_in6.sin6_addr);
+		proxy_addr_in6.sin6_port = htons(host->port);
+		servlen = sizeof(proxy_addr_in6);
+		proxy_addr = (struct sockaddr *) &proxy_addr_in6;
+	} else
+#endif
+	{
+		memset(&proxy_addr_in, 0, sizeof(proxy_addr_in));
+		proxy_addr_in.sin_family = AF_INET;
+		proxy_addr_in.sin_addr.s_addr = inet_addr(host->host->ptr);
+		proxy_addr_in.sin_port = htons(host->port);
+		servlen = sizeof(proxy_addr_in);
+		proxy_addr = (struct sockaddr *) &proxy_addr_in;
+	}
 
-	proxy_addr = (struct sockaddr *) &proxy_addr_in;
 
 	if (-1 == connect(proxy_fd, proxy_addr, servlen)) {
 		if (errno == EINPROGRESS || errno == EALREADY) {
@@ -395,7 +410,7 @@ static int proxy_establish_connection(server *srv, handler_ctx *hctx) {
 	return 0;
 }
 
-void proxy_set_header(connection *con, const char *key, const char *value) {
+static void proxy_set_header(connection *con, const char *key, const char *value) {
     data_string *ds_dst;
 
     if (NULL == (ds_dst = (data_string *)array_get_unused_element(con->request.headers, TYPE_STRING))) {
@@ -407,7 +422,7 @@ void proxy_set_header(connection *con, const char *key, const char *value) {
     array_insert_unique(con->request.headers, (data_unset *)ds_dst);
 }
 
-void proxy_append_header(connection *con, const char *key, const char *value) {
+static void proxy_append_header(connection *con, const char *key, const char *value) {
     data_string *ds_dst;
 
     if (NULL == (ds_dst = (data_string *)array_get_unused_element(con->request.headers, TYPE_STRING))) {
@@ -741,9 +756,19 @@ static handler_t proxy_write_request(server *srv, handler_ctx *hctx) {
 
 	switch(hctx->state) {
 	case PROXY_STATE_INIT:
-		if (-1 == (hctx->fd = socket(AF_INET, SOCK_STREAM, 0))) {
+#if defined(HAVE_IPV6) && defined(HAVE_INET_PTON)
+		if (strstr(host->host->ptr,":")) {
+		    if (-1 == (hctx->fd = socket(AF_INET6, SOCK_STREAM, 0))) {
 			log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed: ", strerror(errno));
 			return HANDLER_ERROR;
+		    }
+		} else
+#endif
+		{
+		    if (-1 == (hctx->fd = socket(AF_INET, SOCK_STREAM, 0))) {
+			log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed: ", strerror(errno));
+			return HANDLER_ERROR;
+		    }
 		}
 		hctx->fde_ndx = -1;
 
@@ -820,17 +845,14 @@ static handler_t proxy_write_request(server *srv, handler_ctx *hctx) {
 
 		chunkqueue_remove_finished_chunks(hctx->wb);
 
-		if (-1 == ret) {
-			if (errno != EAGAIN &&
-			    errno != EINTR) {
-				log_error_write(srv, __FILE__, __LINE__, "ssd", "write failed:", strerror(errno), errno);
+		if (-1 == ret) { /* error on our side */
+			log_error_write(srv, __FILE__, __LINE__, "ssd", "write failed:", strerror(errno), errno);
 
-				return HANDLER_ERROR;
-			} else {
-				fdevent_event_add(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_OUT);
+			return HANDLER_WAIT_FOR_EVENT;
+		} else if (-2 == ret) { /* remote close */
+			log_error_write(srv, __FILE__, __LINE__, "ssd", "write failed, remote connection close:", strerror(errno), errno);
 
-				return HANDLER_WAIT_FOR_EVENT;
-			}
+			return HANDLER_WAIT_FOR_EVENT;
 		}
 
 		if (hctx->wb->bytes_out == hctx->wb->bytes_in) {
@@ -1085,19 +1107,20 @@ static handler_t mod_proxy_check_extension(server *srv, connection *con, void *p
 
 	/* check if extension matches */
 	for (k = 0; k < p->conf.extensions->used; k++) {
+		data_array *ext = NULL;
 		size_t ct_len;
 
-		extension = (data_array *)p->conf.extensions->data[k];
+		ext = (data_array *)p->conf.extensions->data[k];
 
-		if (extension->key->used == 0) continue;
+		if (ext->key->used == 0) continue;
 
-		ct_len = extension->key->used - 1;
+		ct_len = ext->key->used - 1;
 
 		if (s_len < ct_len) continue;
 
 		/* check extension in the form "/proxy_pattern" */
-		if (*(extension->key->ptr) == '/') {
-			if (strncmp(fn->ptr, extension->key->ptr, ct_len) == 0) {
+		if (*(ext->key->ptr) == '/') {
+			if (strncmp(fn->ptr, ext->key->ptr, ct_len) == 0) {
 				if (s_len > ct_len + 1) {
 					char *pi_offset;
 
@@ -1105,15 +1128,17 @@ static handler_t mod_proxy_check_extension(server *srv, connection *con, void *p
 						path_info_offset = pi_offset - fn->ptr;
 					}
 				}
+				extension = ext;
 				break;
 			}
-		} else if (0 == strncmp(fn->ptr + s_len - ct_len, extension->key->ptr, ct_len)) {
+		} else if (0 == strncmp(fn->ptr + s_len - ct_len, ext->key->ptr, ct_len)) {
 			/* check extension in the form ".fcg" */
+			extension = ext;
 			break;
 		}
 	}
 
-	if (k == p->conf.extensions->used) {
+	if (NULL == extension) {
 		return HANDLER_GO_ON;
 	}
 
@@ -1209,7 +1234,7 @@ static handler_t mod_proxy_check_extension(server *srv, connection *con, void *p
 
 		if (ndx >= (int) extension->value->used) {
 			/* didn't found a higher id, wrap to the start */
-			for (ndx = 0; ndx < (int) k; ndx++) {
+			for (ndx = 0; ndx <= (int) k; ndx++) {
 				host = (data_proxy *)extension->value->data[ndx];
 				if (!host->is_disabled) break;
 			}
@@ -1321,6 +1346,7 @@ TRIGGER_FUNC(mod_proxy_trigger) {
 }
 
 
+int mod_proxy_plugin_init(plugin *p);
 int mod_proxy_plugin_init(plugin *p) {
 	p->version      = LIGHTTPD_VERSION_ID;
 	p->name         = buffer_init_string("proxy");
